import time
from datetime import date
from typing import Any, Generator

import pandas as pd
import sqlparse
from rich.progress import Progress

from pysqltools.src.connection import SQLConnection
from pysqltools.src.log import PabLog
from pysqltools.src.queries.query import Query, assign_parameter

lg = PabLog("Insert")


def prepare_value(val: Any) -> Any:
    """
    Format value from Python types to SQL Types
    """
    if isinstance(val, bool):
        return val
    if isinstance(val, dict):
        val = str(val).replace("'", '"')
    if isinstance(val, pd.Timestamp):
        val = "TIMESTAMP '" + str(val) + "'"
    if isinstance(val, date):
        val = "DATE '" + str(val) + "'"
    if isinstance(val, list):
        val = "ARRAY " + str(val)
    if isinstance(val, float):
        val = "DOUBLE '" + str(val) + "'"
    if isinstance(val, int):
        val = "INT '" + str(val) + "'"
    if pd.isnull(val):
        val = "NULL"

    try:
        if (
            "'" in val
            and "DOUBLE" not in val
            and "INT" not in val
            and "TIMESTAMP" not in val
            and "DATE" not in val
        ):
            val = val.replace("'", "''")
    except TypeError:
        lg.log.warning("Not Adding Quotes")

    return val


def join_values(data: list[Any]) -> str:
    """
    Create a String for the VALUES () SQL Syntax
    """
    clean_list = []
    for val in data:
        if isinstance(val, bool):
            val = str(val)
        if (
            isinstance(val, str)
            and "DOUBLE" not in val
            and "INT" not in val
            and "TIMESTAMP" not in val
            and "DATE" not in val
            and "ARRAY" not in val
        ) and val.lower() not in ["true", "false"]:
            val = "'" + val + "'"
        try:
            if "NULL" in val:
                val = "NULL"
        except TypeError:
            lg.log.error("Type not using Nulls")

        clean_list.append(str(val))
    str_data = (", ").join(clean_list)
    return "(" + str_data + ")"


def pandas_to_sql(df: pd.DataFrame) -> Generator[str, None, None]:
    """
    Generator to get one row insert statement
    """
    for row in df.values:
        data_list = [prepare_value(x) for x in row]
        data_string = join_values(data_list)
        yield data_string


def generate_insert_query(
    df: pd.DataFrame, table: str = None, schema: str = None, batch_size: int = 5000
) -> Generator[Query, None, None]:
    if df.empty:
        raise TypeError("DataFrame can not be empty")
    previous_iter = 0
    while previous_iter < len(df):
        percentage = round(100 * previous_iter / len(df), 2)
        lg.log.info("Generating Insert Queries... %s", percentage)
        batch = df.iloc[previous_iter : previous_iter + batch_size]
        data_points = list(pandas_to_sql(batch))
        data_points_string = ",".join(data_points)
        if schema and table:
            table = f"{schema}.{table}"
        if not table:
            table = "{{table}}"
        sql = f"INSERT INTO {table} VALUES {data_points_string}"
        query = Query(sqlparse.format(sql))
        previous_iter += batch_size
        yield query


def insert_pandas(
    df: pd.DataFrame,
    connection: SQLConnection,
    table: str,
    schema: str,
    batch_size: int,
):
    if not table and schema:
        raise TypeError("Table and Schema need to be provided")
    with Progress() as progress:
        iterations = len(df) / batch_size
        task1 = progress.add_task("[red]Generating Queries...", total=1000)
        task2 = progress.add_task("[green]Inserting Data...", total=iterations)
        task3 = progress.add_task("[cyan]Finishing...", total=1000)
        for i in range(1000):
            progress.update(task1, advance=1.0)
        for query in generate_insert_query(df, table, schema, batch_size):
            connection.execute(query)
            progress.update(task2, advance=1)
        for i in range(1000):
            progress.update(task3, advance=1.0)



import sqlite3
from typing import Any, Union

import ibm_db
import mysql
import mysql.connector
import pymssql
import pymysql
import pyodbc
import sqlalchemy
import trino

from pysqltools.src.connection.exceptions import ConnectionException
from pysqltools.src.log import PabLog
from pysqltools.src.queries.query import Query

lg = PabLog("Connections")


class SQLConnection:
    conn = None

    def __init__(
        self,
        conn: Union[
            pymysql.connections.Connection,
            trino.dbapi.Connection,
            pymssql.Connection,
            sqlite3.Connection,
            pyodbc.Connection,
            mysql.connector.connection.MySQLConnection,
            mysql.connector.connection_cext.CMySQLConnection,
            ibm_db.IBM_DBConnection,
            sqlalchemy.Connection,
        ],
    ):
        self.conn = conn

    def execute(self, sql: Query) -> None:
        try:
            if isinstance(self.conn, ibm_db.IBM_DBConnection):
                ibm_db.exec_immediate(self.conn, sql.sql)
            if isinstance(self.conn, sqlalchemy.Connection):
                self.conn.execute(sql.sql)
                self.conn.commit()
            else:
                cursor = self.conn.cursor()
                print(cursor)
                cursor.execute(sql.sql)
                try:
                    self.conn.commit()
                except:
                    lg.log.warning("Connection can not be commited")
        except:
            raise ConnectionException

    def fetch(self, sql: Query):
        try:
            if isinstance(self.conn, ibm_db.IBM_DBConnection):
                stmt = ibm_db.exec_immediate(self.conn, sql.sql)
                rows = []
                row = ibm_db.fetch_assoc(stmt)
                while row:
                    rows.append(row)
                    row = ibm_db.fetch_assoc(stmt)
                return rows

            if isinstance(self.conn, sqlalchemy.engine.base.Connection):
                result = self.conn.execute(sql.sql)
                return result.fetchall()

            else:
                cursor = self.conn.cursor()
                cursor.execute(sql.sql)
                if hasattr(cursor, "fetchall"):
                    return cursor.fetchall()
                else:
                    rows = []
                    row = cursor.fetchone()
                    while row:
                        rows.append(row)
                        row = cursor.fetchone()
                    return rows
        except Exception as e:
            lg.log.error(f"Fetch failed: {e}")
            raise ConnectionException



"""Logger"""

import logging

import pandas as pd
from rich.console import Console
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.table import Table

COLORS = [
    "purple",
    "magenta",
    "turquoise2",
    "blue_violet",
    "orange3",
    "dark_red",
    "yellow4",
    "green",
    "deep_pink3",
    "dark_goldenrod",
    "aquamarine1",
    "chartreuse1",
]


class PabLog:
    """Logger"""

    def __init__(
        self, log_name: str = "Logger", __format: str = "- %(message)s"
    ) -> None:
        """Logger Constructor"""
        self.avaliable_colors = []
        self.rich_handler = RichHandler(rich_tracebacks=True)

        self.avaliable_colors = COLORS

        __handlers = [self.rich_handler]

        logging.basicConfig(format=__format, handlers=__handlers, level=logging.DEBUG)

        self.log = logging.getLogger(log_name)
        self.console = Console()

    def add_table(self, df: pd.DataFrame, title: str = "", max_rows: int = 10) -> None:
        """Add a table in terminal"""
        table = Table(title=title)
        color = 0
        for col in df.columns:
            if color >= len(self.avaliable_colors) - 1:
                color = 0
            else:
                color += 1

            table.add_column(str(col), style=self.avaliable_colors[color])

        for _, row in df.head(max_rows).iterrows():
            table.add_row(*[str(r) for r in row])

        self.console.print(table)

    def add_md(self, markdown: str):
        """Add markdown in terminal"""
        md = Markdown(markdown)
        self.console.print(md)

    def add_title(self, title: str):
        """Add markdown title in terminal"""
        md_ = f" # **{title}** "
        md = Markdown(md_)
        self.console.print(md)



from .src.connection import SQLConnection
from .src.queries.insert import generate_insert_query, insert_pandas
from .src.queries.query import Query, SQLString


def format_sql(sql: str, **kwargs) -> str:
    query = Query(sql=sql).format(kwargs)
    return query.sql



"""
This module contains all the classes and 
functions to interact with SQL text objects.
"""

import datetime
import re
from typing import Generator, Union

import sqlparse
from multimethod import multimethod

from pysqltools.src.queries.exceptions import QueryFormattingError


class SQLString(str):
    """
    String Class used to format queries without adding single quotes.
    """


@multimethod
def assign_parameter(param: str) -> str:
    """
    assign parameter function
    """
    return "'" + param + "'"


@multimethod
def assign_parameter(param: Union[int, float]) -> str:
    """
    assign parameter function
    """
    return param


@multimethod
def assign_parameter(param: SQLString) -> str:
    """
    assign parameter function
    """
    return param


@multimethod
def assign_parameter(param: list[str]) -> str:
    """
    assign parameter function
    """
    string = "("
    for item in param:
        string += f"'{item}', "
    return string[:-2] + ")"


@multimethod
def assign_parameter(param: list[int, float]) -> str:
    """
    assign parameter function
    """
    string = "("
    for item in param:
        string += f"{item}, "
    return string[:-2] + ")"


@multimethod
def assign_parameter(param: datetime.datetime) -> str:
    """
    assign parameter function
    """
    return "datetime '" + param.strftime("%Y-%m-%d %H:%M:%S") + "'"


class Query:
    """
    ### Query Module
    The query module provides a Query class to work with Query objects, which will allow to modify the
    SQL Queries on an easy way with the class methods, and easily access the sql string with the sql
    attribute of the objects.

    To add parameters to the query, use {{parameter}} on the SQL String.

    The current methods are:

    - ctes: Generator that yields the CTEs of the Query
    - selects: Generator that yields the Select statements of the Query
    - Windows: Generator that yields the Window Function contents of the query
    - tables: Generator that yields the detected tables on the query
    - parameters: Generator that yields all the parameters on the Query
    - format: allows to assign values to the parameters in the query. Current supported types are str, int, float, datetime.datetime, list[int, float, str]
    To call the format function, just call the parameters you have defined on your query. Example:
    query:
    `select * from {{table_param}} limit 20`

    function call:
    `query = Query(sql = sql).format(table_param = "MyTable")`
    """

    def __init__(self, sql: str) -> None:
        self.sql = sql.lower()
        self.parsed = sqlparse.parse(sql)[0]

    def ctes(self) -> Generator:
        """
        returns a generator containing all the CTEs on the query
        """
        cte_regex = re.compile(
            r"""(?i)\b(\w+)\s+as\s+\((.*?)\)(?=\s*,|\s*select|\s*insert|\s*update|\s*delete|\s*with|\Z)""",
            re.DOTALL | re.IGNORECASE | re.MULTILINE,
        )

        matches = cte_regex.findall(self.sql)

        for _, match in enumerate(matches, 1):
            cte_name, cte_content = match
            yield (cte_name, cte_content)

    def parameters(self) -> Generator:
        """returns a generator containing all the Parameters on the query.
        Parameters must be between {{ }}"""
        regex = re.compile(r"(?<={{)\S*(?=}})")
        yield from regex.findall(self.sql)

    def __non_greedy_regex(self, keyword_start: str, keyword_end: str) -> Generator:
        """"""
        regex = re.compile(
            rf"(?<={keyword_start}).*?(?={keyword_end})",
            re.DOTALL | re.IGNORECASE | re.MULTILINE,
        )
        yield from [i.strip() for i in regex.findall(self.sql)]

    def selects(self) -> Generator:
        """returns a generator containing all the Select contents on the query"""
        yield from [i.strip() for i in self.__non_greedy_regex("select", "from")]

    def windows(self) -> Generator:
        """returns a generator containing all the Window Functions on the query"""
        yield from [i.strip() for i in self.__non_greedy_regex("over", r"\)")]

    def tables(self) -> Generator:
        """Returns a generator containing all the detected tables"""
        regex = re.compile(
            r"(?<=from|join).*?\s*\S*",
            re.DOTALL | re.IGNORECASE | re.MULTILINE,
        )
        results = regex.findall(self.sql)
        results = [r.strip() for r in results]
        yield from results

    def format(self, **kwargs) -> "Query":
        """
        Allows dynamic variables on SQL Queries.
        The parameters must be between keys i.e. {{parameter}}. Using the format function,
        you can substitute the parameters with python variables.\n
        An special type SQLString is used for tables, as we don't want to include "'" on
        those strings.\n

        `sql = Query(sql="select * from {{my_table}} where country in {{country_list}}")` \n
        `sql.format(my_table=SQLString("schema.table"), country_list = ['ES', 'GB'])`
        """
        try:
            for k, v in kwargs.items():
                self.sql = self.sql.replace("{{" + k + "}}", assign_parameter(v))
            self.parsed = sqlparse.parse(sql=self.sql)[0]
        except:
            raise QueryFormattingError

        return self

    def __str__(self):
        return self.sql

    def __dict__(self):
        return {
            "tables": list(self.tables()),
            "ctes": list(self.ctes),
            "parameters": list(self.parameters()),
        }



"""
Custom Exception for pysqltools
"""


class QueryFormattingError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def formatting_method(fun):
    def inner(*args, **kwargs):
        try:
            print(fun(*args, **kwargs))
            return fun(*args, **kwargs)
        except Exception as e:
            raise QueryFormattingError(str(e))

    return inner
